#include "miniengine.h"
#include <vector>
#include <map>
#include <iterator>

namespace MiniEngine
{
	// namespace global variables
	// globalis valtozok a namespace-en belul
	LPDIRECT3D9				g_pD3D		= NULL;
	LPDIRECT3DDEVICE9		g_pDevice	= NULL;
	D3DPRESENT_PARAMETERS	g_pp;
	LPDIRECT3DSURFACE9		g_pBackBuffer = NULL;
	LPDIRECT3DSURFACE9		g_pDepthBuffer = NULL;
	D3DSURFACE_DESC			g_backBufferDesc;
	LPD3DXEFFECTPOOL		g_pFXPool = NULL;
	bool					g_bRecovering	= false;
	D3DXMATRIX				g_mBackBufferTextureBias;

	// once a resource is loaded (with the first call to the corresponding "Get" function), it's stored in the map.
	// itt tarolodnak a cache-elt resourcek, miutan a resource betoltodott (a megfelelo "Get" fuggveny elso hivasaval).
	typedef std::map<LPCSTR, LPD3DXEFFECT>				tFXMap;
	tFXMap												g_aFX;
	typedef std::map<LPCSTR, Texture>					tTextureMap;
	tTextureMap											g_aTexture;
	typedef std::map<LPCSTR, RenderTarget>				tRTMap;
	tRTMap												g_aRT;

	// these two functions are called internally when needed.
	// ez a ket fuggveny belso hivasra valo, mikor szukseg van ra.
	void CreateUnmanaged()
	{
		g_pDevice->EvictManagedResources();
		for (tRTMap::iterator it=g_aRT.begin();it!=g_aRT.end();it++) (*it).second.Create();
		for (tFXMap::const_iterator it=g_aFX.begin();it!=g_aFX.end();it++) SAFE_ACTION((*it).second, OnResetDevice());
	}

	void DeleteUnmanaged()
	{
		for (tFXMap::const_iterator it=g_aFX.begin();it!=g_aFX.end();it++) SAFE_ACTION((*it).second, OnLostDevice());
		for (tRTMap::iterator it=g_aRT.begin();it!=g_aRT.end();it++) (*it).second.Reset();
	}

// these defines control fullscreen and vsync modes. no support in this for fancy
// requesters, on resolution enumeration. it's up to you to extend it :)
// ezekkel lehet a teljeskepernyos modot illetve a vsync-et ki/be kapcsolni.
// nincs felbontasvalaszto ablak, vagy barmi ilyesmi, ezt mar neked kell belebarkacsolni :)

	//#define FULLSCREEN
	//#define VSYNC

	// initialising d3d. returns false if not successful.
	// a d3d inicializalasa. false-szal ter vissza, ha sikertelen volt.
	bool Init(HWND hWnd)
	{
		if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) ) return false;

		memset(&g_pp, 0, sizeof(g_pp));

		g_pp.BackBufferFormat = D3DFMT_A8R8G8B8;
		g_pp.AutoDepthStencilFormat = D3DFMT_D24S8;
		g_pp.EnableAutoDepthStencil = TRUE;

		RECT wrect;
		GetWindowRect(GetDesktopWindow(), &wrect);

	#ifdef FULLSCREEN
		g_pp.Windowed = FALSE;
		g_pp.BackBufferWidth = wrect.right;//1280;
		g_pp.BackBufferHeight = wrect.bottom;//800;
		#ifdef VSYNC
			g_pp.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
			g_pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		#else
			g_pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
			g_pp.SwapEffect = D3DSWAPEFFECT_COPY;
		#endif
	#else
		g_pp.Windowed = TRUE;
		#ifdef VSYNC
			g_pp.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
			g_pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		#else
			g_pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
			g_pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		#endif
	#endif

		if( FAILED( g_pD3D->CreateDevice(	D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
											D3DCREATE_HARDWARE_VERTEXPROCESSING, &g_pp, &g_pDevice ) ) )
		{
			g_pDevice=NULL;
			SAFE_RELEASE(g_pD3D);
			return false;
		}

		g_pDevice->GetRenderTarget(0, &g_pBackBuffer);
		g_pBackBuffer->GetDesc(&g_backBufferDesc);
		SAFE_RELEASE(g_pBackBuffer);

		D3DXCreateEffectPool(&g_pFXPool);
		CreateUnmanaged();
		return true;
	}

	// terminates the miniengine and uninitialises d3d.
	// kilep a miniengine-bol, es a uninicializalja a d3d-t.
	bool Terminate()
	{
		DeleteUnmanaged();

		for (tFXMap::iterator it=g_aFX.begin();it!=g_aFX.end();it++) SAFE_RELEASE((*it).second);
		for (tTextureMap::iterator it=g_aTexture.begin();it!=g_aTexture.end();it++) ((*it).second).Reset();

		SAFE_RELEASE(g_pFXPool);
		SAFE_RELEASE(g_pDevice);
		SAFE_RELEASE(g_pD3D);
		return true;
	}

	// this function has to be called at the beginning of each frame. it also handles the lost device state,
	// so if it returns with false, don't call anything else (including enddraw!)
	// ezt a fuggvenyt meg kell hivni minden frame elejen. kezeli a lost device allapotot, ugyhogy ha false-szal
	// ter vissza, akkor semmi mas fuggvenyt nem szabad meghivni (beleertve az enddraw-t is!)
	bool BeginDraw()
	{
		if (g_bRecovering)
		{
			if (D3DERR_DEVICENOTRESET==g_pDevice->TestCooperativeLevel())
			{
				g_pDevice->Reset(&g_pp);
				CreateUnmanaged();
				g_bRecovering=false;
			} else {
				return false;
			}
		} else {
			if (D3D_OK!=g_pDevice->TestCooperativeLevel())
			{
				DeleteUnmanaged();
				g_bRecovering=true;
				return false;
			}
		}

		g_pDevice->GetRenderTarget(0, &g_pBackBuffer);
		g_pDevice->GetDepthStencilSurface(&g_pDepthBuffer);
		g_pBackBuffer->GetDesc(&g_backBufferDesc);

		g_mBackBufferTextureBias=D3DXMATRIX(	0.5f, 0.0f, 0.0f, 0.0f,
									0.0f, -0.5f, 0.0f, 0.0f,
									0.0f, 0.0f, 1.0f, 0.0f,
									0.5f+0.5f/(float)g_backBufferDesc.Width, 0.5f+0.5f/(float)g_backBufferDesc.Height, 0.0f, 1.0f);

		g_pDevice->BeginScene();

		return true;
	}

	// call this function to end rendering for the actual frame.
	// a fuggveny hivasa lezarja az aktualis frame rendereleset.
	bool EndDraw()
	{
		g_pDevice->EndScene();
		g_pDevice->Present(NULL, NULL, NULL, NULL);
		SAFE_RELEASE(g_pBackBuffer);
		SAFE_RELEASE(g_pDepthBuffer);
		return true;
	}

	// get functions
	const LPDIRECT3D9& GetD3D() { return g_pD3D; }
	const LPDIRECT3DDEVICE9& GetDevice() { return g_pDevice; }
	const LPDIRECT3DSURFACE9& GetBackBuffer() { return g_pBackBuffer; }
	const LPDIRECT3DSURFACE9& GetDepthBuffer() { return g_pDepthBuffer; }
	const D3DSURFACE_DESC& GetBackBufferDesc() { return g_backBufferDesc; }
	const D3DXMATRIX& GetBackBufferTextureBias() { return g_mBackBufferTextureBias; }

	RenderTarget::RenderTarget() :
		m_pTexture(NULL),
		m_pSurface(NULL)
	{
		memset(&m_surfaceDesc, 0, sizeof(D3DSURFACE_DESC));
		D3DXMatrixIdentity(&m_mTextureBias);
	}

	RenderTarget::~RenderTarget()
	{
		Reset();
	}

	void RenderTarget::Reset()
	{
		SAFE_RELEASE(m_pSurface);
		SAFE_RELEASE(m_pTexture);
	}

	void RenderTarget::Create(const D3DSURFACE_DESC& surfaceDesc, u_int uiCloneWidthDivider, u_int uiCloneHeightDivider)
	{
		m_surfaceDesc=surfaceDesc;
		m_uiCloneWidthDivider=uiCloneWidthDivider;
		m_uiCloneHeightDivider=uiCloneHeightDivider;
		Create();
	}

	void RenderTarget::Create()
	{
		Reset();
		if (m_uiCloneWidthDivider) m_surfaceDesc.Width=g_backBufferDesc.Width/m_uiCloneWidthDivider;
		if (m_uiCloneHeightDivider)	m_surfaceDesc.Height=g_backBufferDesc.Height/m_uiCloneHeightDivider;

		if ((D3DFMT_D24X8==m_surfaceDesc.Format) || (D3DFMT_D24S8==m_surfaceDesc.Format) || (D3DFMT_D16==m_surfaceDesc.Format))
		{
			g_pDevice->CreateDepthStencilSurface(m_surfaceDesc.Width, m_surfaceDesc.Height, m_surfaceDesc.Format, m_surfaceDesc.MultiSampleType, m_surfaceDesc.MultiSampleQuality, true, &m_pSurface, NULL);
		} else {
			g_pDevice->CreateTexture(m_surfaceDesc.Width, m_surfaceDesc.Height, 1, m_surfaceDesc.Usage, m_surfaceDesc.Format, m_surfaceDesc.Pool, &m_pTexture, NULL);
			m_pTexture->GetSurfaceLevel(0, &m_pSurface);
			g_pDevice->ColorFill(m_pSurface, NULL, 0xffffffff);
		}

		m_mTextureBias=D3DXMATRIX(	0.5f, 0.0f, 0.0f, 0.0f,
									0.0f, -0.5f, 0.0f, 0.0f,
									0.0f, 0.0f, 1.0f, 0.0f,
									0.5f+0.5f/(float)m_surfaceDesc.Width, 0.5f+0.5f/(float)m_surfaceDesc.Height, 0.0f, 1.0f);
	}

	const Texture& GetTexture(LPCSTR sTextureName)
	{
		Texture &pTexture = g_aTexture[sTextureName];
		if (!pTexture.GetTexture()) pTexture.Create(sTextureName);
		return pTexture;
	}

	const LPD3DXEFFECT& GetFX(LPCSTR sFXName)
	{
		LPD3DXEFFECT &pFX = g_aFX[sFXName];
		if (!pFX) D3DXCreateEffectFromFile(g_pDevice, sFXName, NULL, NULL, D3DXSHADER_PACKMATRIX_ROWMAJOR, g_pFXPool, &pFX, NULL);
		return pFX;
	}

	const RenderTarget& GetRT(LPCSTR sRTName, int iWidth, int iHeight, D3DFORMAT surfaceFormat)
	{
		RenderTarget& renderTarget = g_aRT[sRTName];
		if (!renderTarget.GetSurface())
		{
			D3DSURFACE_DESC surfaceDesc;
			surfaceDesc.Width=(u_int)iWidth;
			surfaceDesc.Height=(u_int)iHeight;
			surfaceDesc.Format=surfaceFormat;
			surfaceDesc.Usage=D3DUSAGE_RENDERTARGET;
			surfaceDesc.Pool=D3DPOOL_DEFAULT;
			surfaceDesc.MultiSampleType=D3DMULTISAMPLE_NONE;
			surfaceDesc.MultiSampleQuality=0;
			surfaceDesc.Type=D3DRTYPE_TEXTURE;
			renderTarget.Create(surfaceDesc, iWidth<0 ? (u_int)-iWidth : 0, iHeight<0 ? (u_int)-iHeight : 0);
		}
		return renderTarget;
	}

	void DrawWithFX(const LPD3DXEFFECT& pFX, D3DXHANDLE hTechnique, u_int uiNumFaces)
	{
		pFX->SetTechnique(hTechnique);

		UINT uiNumPasses;
		pFX->Begin(&uiNumPasses, 0);
		for (UINT i=0;i<uiNumPasses;i++)
		{
			pFX->BeginPass(i);
			if (uiNumFaces>0)
			{
				g_pDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, uiNumFaces);
			}
			pFX->EndPass();
		}
		pFX->End();
	}

	void DrawIndexedWithFX(const LPD3DXEFFECT& pFX, D3DXHANDLE hTechnique, u_int uiNumVertices, u_int uiNumFaces)
	{
		pFX->SetTechnique(hTechnique);

		UINT uiNumPasses;
		pFX->Begin(&uiNumPasses, 0);
		for (UINT i=0;i<uiNumPasses;i++)
		{
			pFX->BeginPass(i);
			if (uiNumVertices>0 && uiNumFaces>0)
			{
				g_pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, uiNumVertices, 0, uiNumFaces);
			}
			pFX->EndPass();
		}
		pFX->End();
	}

	void DrawUPWithFX(const LPD3DXEFFECT& pFX, D3DXHANDLE hTechnique, void* pVB, u_int uiStride, void *pIB, u_int uiNumVertices, u_int uiNumFaces)
	{
		pFX->SetTechnique(hTechnique);

		UINT uiNumPasses;
		pFX->Begin(&uiNumPasses, 0);
		for (UINT i=0;i<uiNumPasses;i++)
		{
			pFX->BeginPass(i);
			if (uiNumVertices>0 && uiNumFaces>0)
			{
				g_pDevice->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, uiNumVertices, uiNumFaces, pIB, D3DFMT_INDEX16, pVB, uiStride);
			}
			pFX->EndPass();
		}
		pFX->End();
	}

	void DrawMeshWithFX(const LPD3DXEFFECT& pFX, D3DXHANDLE hTechnique, const LPD3DXMESH& pMesh)
	{
		pFX->SetTechnique(hTechnique);

		UINT uiNumPasses;
		pFX->Begin(&uiNumPasses, 0);
		for (UINT i=0;i<uiNumPasses;i++)
		{
			pFX->BeginPass(i);
			if (pMesh)
			{
				pMesh->DrawSubset(0);
			}
			pFX->EndPass();
		}
		pFX->End();
	}

	void Reflection(const D3DXMATRIX& mView, const D3DXMATRIX& mProj, const D3DXPLANE& Plane, D3DXMATRIX& mNewView, D3DXMATRIX& mNewProj)
	{
		D3DXMATRIX mReflection;
		D3DXMatrixReflect(&mReflection, &Plane);
		mNewView=mReflection*mView;

		D3DXPLANE ViewReflectionPlane;
		D3DXMATRIX mInverseTransposeView;
		D3DXMatrixInverse(&mInverseTransposeView, NULL, &mNewView);
		D3DXMatrixTranspose(&mInverseTransposeView, &mInverseTransposeView);
		D3DXPlaneTransform(&ViewReflectionPlane, &Plane, &mInverseTransposeView);
		MatrixOblique(&mNewProj, &ViewReflectionPlane, &mProj);
	}

	// lengyel's implementation, adjusted for d3dx structures.
	void MatrixOblique(D3DXMATRIX* pOutProjection, const D3DXPLANE* pPlane, const D3DXMATRIX* pProjection)
	{
		D3DXPLANE clipPlane;
		D3DXPlaneNormalize(&clipPlane, pPlane);

		D3DXMATRIX mInvTrans;
		D3DXMatrixInverse(&mInvTrans, NULL, pProjection);
		D3DXMatrixTranspose(&mInvTrans, &mInvTrans);

		D3DXPlaneTransform(&clipPlane, &clipPlane, &mInvTrans);

		if (clipPlane.d == 0.0f)  // or less than a really small value
		{
			// plane is perpendicular to the near plane
			*pOutProjection=*pProjection;
			return;
		}

		D3DXPlaneNormalize(&clipPlane, &clipPlane);  //  bug fix by David Whatley
		if (clipPlane.d > 0.0f)
		{
			clipPlane = -clipPlane;
			clipPlane.d += 1;
		} 

		D3DXMATRIX mClipProj;
		D3DXMatrixIdentity(&mClipProj);
		mClipProj(0,2) = clipPlane.a;
		mClipProj(1,2) = clipPlane.b;
		mClipProj(2,2) = clipPlane.c;
		mClipProj(3,2) = clipPlane.d;
		*pOutProjection=*pProjection*mClipProj;
	}

	Texture::Texture() :
		m_pTexture(NULL),
		m_vDimensions(0, 0, 0, 0)
	{
	}

	Texture::~Texture()
	{
		Reset();
	}

	void Texture::Reset()
	{
		SAFE_RELEASE(m_pTexture);
		m_vDimensions=D3DXVECTOR4(0, 0, 0, 0);
	}

	void Texture::Create(LPCSTR sFileName)
	{
		Reset();
		D3DXIMAGE_INFO info;
		if (D3D_OK==D3DXGetImageInfoFromFile(sFileName, &info))
		{
			m_vDimensions=D3DXVECTOR4((float)info.Width, (float)info.Height, (float)info.Depth, 0);
			switch (info.ResourceType)
			{
				case D3DRTYPE_TEXTURE:
				{
					LPDIRECT3DTEXTURE9 pTexture;
					D3DXCreateTextureFromFile(g_pDevice, sFileName, &pTexture);
					m_pTexture=pTexture;
				} break;

				case D3DRTYPE_VOLUMETEXTURE:
				{
					LPDIRECT3DVOLUMETEXTURE9 pTexture;
					D3DXCreateVolumeTextureFromFileEx(g_pDevice, sFileName, info.Width, info.Height, info.Depth, info.MipLevels, 0, D3DFMT_FROM_FILE, D3DPOOL_MANAGED, D3DX_DEFAULT, D3DX_DEFAULT, 0, NULL, NULL, &pTexture);
					m_pTexture=pTexture;
				} break;

				case D3DRTYPE_CUBETEXTURE:
				{
					LPDIRECT3DCUBETEXTURE9 pTexture;
					D3DXCreateCubeTextureFromFile(g_pDevice, sFileName, &pTexture);
					m_pTexture=pTexture;
				} break;

				default:;
			};
		}
	}
}